"""GithubRepo.py module
"""

from thipster.repository.JSONRepo import JSONRepo
import requests


class GithubRepo(JSONRepo):
    """Class representing a GitHub resources Repository

    JSON Models of resources and services offered by supported cloud providers are
    stored in a repository.
    This class is used to access those models if they are located in a GitHub repo.
    """

    def __init__(self, repo: str) -> None:
        """
        Parameters
        ----------
        repo : str
            Name of the repository, for example : 'THipster/models'
        """
        super().__init__()
        self.__repo = repo

    def get_json(self, name: str) -> str | bytes | bytearray:
        """Method to get the json file from the GitHub repository

        Parameters
        ----------
        name : str
            Name of the desired resource

        Returns
        -------
        str | bytes | bytearray
            Content of the JSON file defining the desired resource
        """
        response = requests.get(
            f'https://raw.githubusercontent.com/{self.__repo}/main/{name}.json',
        )

        return response.content



import os
from thipster.engine.I_Parser import I_Parser
from thipster.engine.ParsedFile import ParsedFile
from thipster.parser.YAMLParser import YAMLParser
from thipster.parser.dsl_parser.DSLParser import DSLParser


class ParserPathNotFound(Exception):
    def __init__(self, path, *args: object) -> None:
        super().__init__(*args)

        self.__message = f'Path not found : {path}'

    @property
    def message(self):
        return self.__message


class ParserFactory(I_Parser):

    def __getfiles(self, path: str) -> list[str]:
        """Recursively get all files names in the requested directory and its\
              sudirectories
        Can be run on a path file aswell

        Parameters
        ----------
        path: str
            Path to run this function into

        Returns
        -------
        list[str]
            A list of all the filenames 
        """

        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise ParserPathNotFound(path)

        files = []

        if os.path.isdir(path):
            for content in os.listdir(path):
                files += self.__getfiles(f'{path}/{content}')

        if os.path.isfile(path):
            return [path]

        return files

    def __yamlParser(self) -> YAMLParser:
        return YAMLParser

    def __dslParser(self) -> DSLParser:
        return DSLParser

    def __noParser(self):
        raise Exception()

    def run(self, path: str) -> ParsedFile:
        """Run the ParserFactory

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        files = self.__getfiles(path)

        res = ParsedFile()
        for file in files:
            parsedFile = self.__getParser(file).run(file)
            res.resources += parsedFile.resources

        return res

    def __getParser(self, path) -> I_Parser:
        __parsers = {
            '.yaml': self.__yamlParser,
            '.yml': self.__yamlParser,
            '.jinja': self.__yamlParser,
            '.thips': self.__dslParser,
        }

        _, pathExtension = os.path.splitext(path)

        return __parsers.get(pathExtension, self.__noParser)()



from abc import ABC, abstractmethod
import thipster.engine.ResourceModel as rm
import thipster.engine.ParsedFile as pf


class I_Terraform(ABC):
    """Terraform module interface

    Methods
    -------
    generate()
        Generates Terraform code from parsed file and models
    plan()
        Get plan from generated terraform code
    apply()
        Apply generated terraform code

    """
    @abstractmethod
    def apply(self):
        """Apply generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement apply()')

    @abstractmethod
    def generate(self, file: pf.ParsedFile, models: dict[str, rm.ResourceModel]):
        """Generates Terraform code from parsed file and models

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement generate()')

    @abstractmethod
    def init(self):
        """Init Terraform for generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement generate()')

    @abstractmethod
    def plan(self):
        """Get plan from generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement plan()')



import google.auth
from thipster.engine.I_Auth import I_Auth
from cdktf_cdktf_provider_google.provider import GoogleProvider


class GoogleAuth(I_Auth):
    """Authentication to GCP Projects

    Methods
    -------
    authenticate(app: Construct)
        Generates the google provider in terraform

    """

    def __init__(self) -> None:
        return GoogleAuth

    def authenticate(app):
        """Generates the google provider block for the Terraform CDK

        Parameters
        ----------
        app: Construct
            CDK Construct where the provider is created

        """

        credentials, project_id = google.auth.default()

        GoogleProvider(
            app, "default_google",
            project=project_id,
            access_token=credentials.token,
        )



import os
import shutil
from cdktf_cdktf_provider_google.provider import GoogleProvider

from thipster.engine.Engine import Engine
from thipster.engine.I_Auth import I_Auth
from thipster.parser.ParserFactory import ParserFactory
import pytest
from thipster.repository.LocalRepo import LocalRepo
import thipster.terraform.CDK as cdk

LOCAL_REPO = 'tests/resources/e2e/models'
REMOTE_REPO = 'THipster/models'


class MockAuth(I_Auth):
    def authenticate(app):
        GoogleProvider(
            app, "default_google",
            project="project_id",
            credentials="credentials.token",
        )


def create_dir(dirname: str, files: dict[str, str]):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)

    dirname = os.path.abspath(dirname)
    for name, content in files.items():
        create_file(name, content, dirname)

    def destroy_files():
        shutil.rmtree(dirname)

    return destroy_files


def create_file(filename: str, content: str, dirname: str = 'test'):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    dirname = os.path.abspath(dirname)

    file = open(f'{dirname}/{filename}', 'w')
    file.write(content)
    file.close()


def __test_file(file: str, local_repo: str = LOCAL_REPO):

    path_input = 'test'
    __destroy_dir = create_dir(
        path_input,
        {
            'test_file.thips':
            file,
        },
    )

    engine = Engine(
        ParserFactory(),
        LocalRepo(local_repo),
        MockAuth,
        cdk.CDK(),
    )
    try:
        output = engine.run(path_input)
    except Exception as e:
        raise e
    finally:
        __destroy_dir()

        if os.path.exists('cdktf.out'):
            shutil.rmtree('cdktf.out')

    return output


def test_bucket():
    out = __test_file(
        file="""
bucket my-bucket:
\tregion : euw
    """,
    )

    assert isinstance(out[0], list)
    assert len(out[0]) == 1

    assert out[0][0] == 'cdktf.out/stacks/thipster_infrastructure'


def test_empty_bucket():
    out = __test_file(
        file="""
bucket dzvhvzarbazkhr:

    """,
    )

    assert isinstance(out[0], list)
    assert len(out[0]) == 1

    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]

    out = __test_file(
        file="""
bucket dzvhvzarbazkhr:

bucket ezezeaz:
    region: europe
    """,
    )

    assert isinstance(out[0], list)
    assert len(out[0]) == 1

    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]


def test_dep_with_no_options():
    with pytest.raises(cdk.CDKMissingAttributeInDependency):
        __test_file(
            file="""
bucket_bad_dep_parent my-bucket:
\tregion : euw
        """,
        )


def test_cyclic_deps():
    with pytest.raises(cdk.CDKCyclicDependencies):
        __test_file(
            file="""
bucket_bad_dep_cyclic my-bucket:
\tregion : euw
        """,
        )


def test_lb():
    out = __test_file(
        file="""
network lb-net:

subnetwork lb-subnet:
\tregion: europe-west1b
\tip_range: 10.0.1.0/24

loadbalancer my-lb:
\tload_balancing_scheme: EXTERNAL
    """,
    )

    assert isinstance(out[0], list)
    assert len(out[0]) == 1

    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]


def test_lb_single_file():
    out = __test_file(
        file="""
network lb-net:

subnetwork lb-subnet:
\tregion: europe-west1b
\tip_range: 10.0.1.0/24

loadbalancer my-lb:
\tload_balancing_scheme: EXTERNAL
    """,
    )

    assert isinstance(out[0], list)
    assert len(out[0]) == 1

    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]


def test_internal_object():
    out = __test_file(
        file="""
firewall testParent:
\tdirection: EGRESS
        """,
    )
    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]

    out = __test_file(
        file="""
firewall testParent:
\tdirection: EGRESS

\tallow:
\t\tprotocol: http
        """,
    )
    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]


def test_bucket_cors():
    out = __test_file(
        file="""
bucket corsBucket:
    cors:
        origin:
            - "http://image-store.com"
        method:
            - "*"
        responseHeader:
            - "*"
        maxAge: 400
        """,
    )
    assert 'cdktf.out/stacks/thipster_infrastructure' in out[0]



"""_Engine.py module.
"""
import time

from thipster.engine.I_Parser import I_Parser
from thipster.engine.I_Repository import I_Repository
from thipster.engine.I_Auth import I_Auth
from thipster.engine.I_Terraform import I_Terraform
import thipster.engine.ParsedFile as pf
from thipster.engine.ResourceModel import ResourceModel


class Engine():
    """The engine of thipster

    The core of the application, it is used to call and link all
    interfaces together.

    Methods
    -------
    run(filename: str)
        Processes a fileName and creates the corresponding Cloud architecture plan

    """

    importedPackages = []

    def __init__(
            self, parser: I_Parser,
            repository: I_Repository,
            auth: I_Auth,
            terraform:  I_Terraform,
    ):
        """
        Parameters
        ----------
        parser : I_Parser
            Instance of a Parser class
        repository : I_Repository
            Instance of a Respository class
        auth : I_Auth
            Instance of an Auth class
        terraform : I_Terraform
            Instance of a Terraform class

        """
        self.__parser = parser
        self.__repository = repository
        self.__auth = auth
        self.__terraform = terraform

    def run(self, path: str) -> tuple[list[str], str]:
        """Returns json Terraform files from the input file name

        Calls the different run methods of the parser, repository,
        auth and terraform modules.
        Transforms the inputed filename into a Cloud architecture plan.

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        tuple[list[str], str]
            A tuple made up of the list of directories containing the Terraform json 
            files and a string with the results of the Terraform plan
        """
        # Parse file or directory
        file = self._parse_files(path)[0]

        # Get needed models
        models = self._get_models(file)[0]

        # Generate Terraform files
        dirs = self.__terraform.generate(file, models, self.__auth)

        self.__terraform.init()

        return dirs, self.__terraform.plan()

    def _parse_files(self, path: str) -> tuple[pf.ParsedFile, float]:
        """Parse the input file or directory

        Parameters
        ----------
        path : str
            The path of the files to be processed

        Returns
        -------
        tuple[pf.ParsedFile, float]
            A tuple made up of the ParsedFile object and the time it took to parse it
        """
        start = time.time()

        file = self.__parser.run(path)
        assert type(file) == pf.ParsedFile

        end = time.time()
        return file, end - start

    def _get_models(
        self, file: pf.ParsedFile,
    ) -> tuple[dict[str, ResourceModel], float]:
        """Get the models from the repository

        Parameters
        ----------
        file : pf.ParsedFile
            The ParsedFile object containing the resources defined in the input file

        Returns
        -------
        tuple[dict[str, ResourceModel], float]
            A tuple made up of the dictionary of models and the time it took to get them
        """
        start = time.time()
        types = [r.type for r in file.resources]
        models = self.__repository.get(types)
        end = time.time()

        return models, end - start



import copy
import shutil
import subprocess
import sys
import os
import importlib
import uuid
from thipster.engine.I_Auth import I_Auth

from python_terraform import Terraform
from constructs import Construct
from cdktf import App, TerraformStack, TerraformOutput

import thipster.engine.ResourceModel as rm
import thipster.engine.ParsedFile as pf
from thipster.engine.I_Terraform import I_Terraform
from thipster.helpers import createLogger as Logger


class CDKException(Exception):
    @property
    def message(self):
        return str(self)


class CDKInvalidAttribute(CDKException):
    def __init__(self, attr: str, modelType: str, **args: object) -> None:
        super().__init__(*args)
        self.__attr = attr
        self.__modelType = modelType

    def __str__(self) -> str:
        return f'{self.__attr} in {self.__modelType} but not useful'


class CDKMissingAttribute(CDKException):
    pass


class CDKMissingAttributeInDependency(CDKMissingAttribute):
    pass


class CDKDependencyNotDeclared(CDKException):
    def __init__(self, depType: str, depName: str, **args: object) -> None:
        super().__init__(*args)
        self.__name = depName
        self.__type = depType

    def __str__(self) -> str:
        return f'{self.__type} {self.__name} not declared \
(be sure to declare it before using it)'


class CDKCyclicDependencies(CDKException):
    def __init__(self, stack: list[str], **args: object) -> None:
        super().__init__(*args)
        self.__stack = stack

    def __str__(self) -> str:
        return ','.join(self.__stack)


class CDK(I_Terraform):
    _models = []
    _parentStack = []
    _importedPackages = []

    _created_resources = {}
    _logger = Logger(__name__)

    def apply(self):
        """Apply generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        t = Terraform()
        _, stdout, stderr = t.apply()
        return stdout + stderr

    def generate(
        self, file: pf.ParsedFile, models: dict[str, rm.ResourceModel],
        _auth: I_Auth,
    ):

        CDK._models = models
        # Init CDK
        app = App()

        f_position = file.resources[0].position
        file_name = f_position.fileName if f_position else 'thipster_infrastructure'

        CDK._logger.debug('Creating tf code for file %s', file_name)

        # Declare new stack in CDK
        class _ResourceStack(TerraformStack):
            def __init__(self, scope: Construct, ns: str):
                super().__init__(scope, ns)

                _auth.authenticate(self)

                for resource in file.resources:
                    res = CDK._create_resource_from_resource(
                        self,
                        resource=resource,
                    )

                    CDK._created_resources[f'{resource.type}/{resource.name}'] = res.id

                    TerraformOutput(
                        self, f"{resource.name}_id",
                        value=res.id,
                    )

        _ResourceStack(app, file_name)

        # Synth files
        app.synth()

        self.__dirs = [
            f'{app.outdir}/stacks/{c.node.path}' for c in app.node.children
        ]

        CDK._logger.info('Created %s terraform file(s)', len(self.__dirs))

        # Move files
        for dirname in self.__dirs:
            shutil.move(
                os.path.join(os.getcwd(), dirname, "cdk.tf.json"),
                os.path.join(os.getcwd(), "thipster.tf.json"),
            )

        # Delete cdktf.out directory
            for content in os.listdir(os.path.join(os.getcwd(), dirname)):
                os.remove(f'{dirname}/{content}')
            os.rmdir(dirname)
        for content in os.listdir(os.path.join(os.getcwd(), 'cdktf.out')):
            d = f'cdktf.out/{content}'
            os.rmdir(d) if os.path.isdir(d) else os.remove(d)
        os.rmdir('cdktf.out')

        return self.__dirs

    def init(self):
        """Get plan from generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        t = Terraform()
        _, stdout, stderr = t.init()
        return stdout + stderr

    def plan(self):
        """Get plan from generated terraform code

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        t = Terraform()
        _, stdout, stderr = t.plan(out="thipster.tfplan")
        return stdout + stderr

    def _pip_install(package: str):
        if package not in CDK._importedPackages:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'install', '-qqq', package],
            )
            CDK._importedPackages.append(package)

    def _import(packageName: str, moduleName: str, className: str) -> type:

        module = importlib.import_module(f'{packageName}.{moduleName}')
        class_ = getattr(module, className)

        return class_

    def _create_default_resource(
        self, typ: str, parentName: str | None = None,
        noModif: bool = True, noDependencies: bool = False,
    ):
        # Checks for cyclic dependency
        if typ in CDK._parentStack:
            CDK._logger.error('%s already present in parent Stack', typ)
            raise CDKCyclicDependencies(CDK._parentStack)

        CDK._parentStack.append(typ)

        model = CDK._models[typ]

        # Checks that all attributes are optional
        if noModif and not all(map(lambda x: x.optional, model.attributes.values())):
            raise CDKMissingAttributeInDependency(typ)

        # Import package and class
        CDK._pip_install(model.cdk_provider)
        resourceClass = CDK._import(
            model.cdk_provider, model.cdk_module, model.cdk_name,
        )

        # Process name + default values
        deps = copy.deepcopy(model.dependencies)
        resource_args = {}

        name = f"{parentName}-{uuid.uuid4()}"
        if model.name_key:
            resource_args[model.name_key] = name

        for _, v in model.attributes.items():
            resource_args[v.cdk_name] = v.default

        # Create default defendencies if needed
        if not noDependencies:
            CDK._create_dependencies(self, deps, resource_args, name)

        CDK._logger.debug('Created default %s named %s', resourceClass, name)

        return (resourceClass, name, resource_args)

    def _create_resource_from_args(
        self, typ: str, args: list[pf.ParsedAttribute] | None,
    ):
        resourceClass, resourceName, resource_args = CDK._create_default_resource(
            self,
            typ,
            noModif=False,
            noDependencies=True,
        )
        model = CDK._models[typ]

        # Process attributes
        deps = copy.deepcopy(model.dependencies)
        for attribute in args:
            if attribute.name and model.name_key:
                resource_args[model.name_key] = attribute.name

            resource_args, deps = CDK._process_attribute(
                self,
                model=model,
                attribute=attribute,
                resource_args=resource_args,
                deps=deps,
            )

        # Create missing dependencies
        CDK._create_dependencies(self, deps, resource_args, resourceName)

        # Create resource
        CDK._parentStack.remove(typ)
        if not model.name_key:
            return resourceClass(**resource_args)

        CDK._logger.debug(
            'Created default %s named %s',
            resourceClass, resourceName,
        )
        return resourceClass(self, resourceName, **resource_args)

    def _create_resource_from_resource(self, resource: pf.ParsedResource):
        # Create resource with default values
        resourceClass, _, resource_args = CDK._create_default_resource(
            self,
            resource.type,
            noModif=False,
            noDependencies=True,
        )
        model = CDK._models[resource.type]

        if model.name_key:
            resource_args[model.name_key] = resource.name

        # Process attributes
        deps = copy.deepcopy(model.dependencies)
        for attribute in resource.attributes:
            resource_args, deps = CDK._process_attribute(
                self,
                model=model,
                attribute=attribute,
                resource_args=resource_args,
                deps=deps,
            )

        # Create missing dependencies
        CDK._create_dependencies(self, deps, resource_args, resource.name)

        # Create resource
        CDK._parentStack.remove(resource.type)

        CDK._logger.debug(
            'Created resource %s named %s',
            resourceClass, resource.name,
        )

        return resourceClass(self, resource.name, **resource_args)

    def _process_attribute(
        self,
        model: rm.ResourceModel,
        attribute: pf.ParsedAttribute,
        resource_args: dict[str, object],
        deps: dict[str, dict[str, object]] | None,
    ):
        # Checks if attribute is an explicit dependency
        if attribute.name in deps:
            created_name = f"{deps[attribute.name]['resource']}/{attribute.value}"

            if created_name not in CDK._created_resources:
                raise CDKDependencyNotDeclared(attribute.name, attribute.value)

            resource_args[attribute.name] = CDK._created_resources[created_name]
            del deps[attribute.name]
            return resource_args, deps

        # Checks if attribute is an internal object
        if attribute.name in model.internalObjects:
            resource_args = CDK._handle_internal_object(
                self,
                model=model,
                attribute=attribute,
                resource_args=resource_args,
            )
            return resource_args, deps

        # Checks if attribute is expected as an attibute
        if attribute.name not in model.attributes:
            raise CDKInvalidAttribute(attribute.name, model.type)

        # Processes list attribute
        attribute_value = attribute.value
        if model.attributes[attribute.name].is_list:
            if type(attribute.value) is list:
                attribute_value = [i.value for i in attribute.value]
            else:
                attribute_value = [attribute.value]

        # Sets attribute value
        resource_args[model.attributes[attribute.name].cdk_name] = attribute_value

        return resource_args, deps

    def _handle_internal_object(
        self,
        model: rm.ResourceModel,
        attribute: pf.ParsedAttribute,
        resource_args: dict,
    ):
        res = CDK._create_resource_from_args(
            self,
            model.internalObjects[attribute.name]['resource'],
            attribute.value,
        )

        int_obj = model.internalObjects[attribute.name]
        var_type = int_obj['var_type'] if 'var_type' in int_obj else 'Unknown'

        if 'list' in var_type:
            if attribute.name not in resource_args:
                resource_args[attribute.name] = []
            resource_args[attribute.name] += [res]
        else:
            resource_args[attribute.name] = res

        return resource_args

    def _create_dependencies(self, deps, resource_args, name):
        for k, d in deps.items():
            c, n, a = CDK._create_default_resource(self, d['resource'], name)
            id = c(self, n, **a).id
            if k:
                resource_args[k] = id

            CDK._parentStack.remove(d['resource'])



from abc import ABC, abstractmethod
from thipster.engine.ParsedFile import ParsedFile


class I_Parser(ABC):
    """Parser module interface

    Methods
    -------
    run()
        Abstract run method

    """
    @abstractmethod
    def run(self, path: str) -> ParsedFile:
        """Abstract run method

        Returns
        -------
        ParsedFile
            Object containing a representation of an input file broken down into
            individual resources

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement run()')



from abc import ABC, abstractmethod
from thipster.engine.ResourceModel import ResourceModel


class I_Repository(ABC):
    """Repository module interface

    Methods
    -------
    get()
        Abstract get method

    """
    @abstractmethod
    def get(self, resourceNames: list[str]) -> dict[str, ResourceModel]:
        pass



from setuptools import setup, find_packages

__version__ = '0.13.6'

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md") as rm:
    readme = rm.read()

setup(
    name="thipster",
    authors=[
        {"name": "rcattin", "email": "rafa.cattin+thipster@gmail.com"},
        {"name": "gsuquet", "email": "gsuquet@ippon.fr"},
    ],
    description="",
    long_description=readme,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python",
    ],

    download_url="https://github.com/THipster/THipster.git",
    url="https://github.com/THipster/THipster",

    version=__version__,
    install_requires=required,
    packages=find_packages(
        exclude=['pipelines'],
    ),
    extras_require={
        'test': [
            'pytest',
            'pytest-mock',
        ],
        'dev': [
            'pytest',
            'pytest-mock',
            'dagger.io',
            'pre-commit',
        ],
        'doc': [
            'sphinx',
        ],
    },
)



class DSLParserBaseException(Exception):
    def __init__(self, message, *args: object) -> None:
        super().__init__(*args)

        self.__message = message

    @property
    def message(self):
        return self.__message


class DSLParserPathNotFound(DSLParserBaseException):
    def __init__(self, file, *args: object) -> None:
        super().__init__(f'Path not found : {file}', *args)
